/* -*- mode: C++; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/* Copyright (C) 2010, Pascal Giorgi
 *
 * Written by Pascal Giorgi pascal.giorgi@lirmm.fr
 *
 * ========LICENCE========
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * ========LICENCE========
 */



#include <iostream>
#include <vector>
#include <string>
#include <iterator>
#include <algorithm> 
#include <cstdlib>
#include <ctime>
#include <iomanip>

#include "timing.h"
#include "fft-polmul.h"
#include "chebyshev-mul.h"

using namespace std;

#define MAX_V 50 
 

double  rand_bound(size_t N){return  (2.*N* (rand() / (double)RAND_MAX)) - (double)N;}



double t_start,t_end, tps;
#define BENCH(__N__ , __LINE__ , __CMD__)			\
	t_start = microseconds();				\
	for (unsigned long  cmpt = 0; cmpt < __N__; ++cmpt) {	\
		__CMD__						\
			}					\
	t_end = microseconds();					\
	tps = t_end - t_start;					\
	tps /= (double)(__N__);					\
	cout<<"Timing of "<<__LINE__<<" : "<<tps<<endl;

template<typename T>
void bench2(void mul(vector<T>&,const vector<T>&, const vector<T>&), vector<T> &C, vector<T> A, vector<T> B, size_t loop, string print){
	double t1,t2,t;
	t=0.0;
	double check=0.0;
	for (size_t i= 0; i< loop; i++) 
		{		
			t1 = get_time(); 
			mul(C,A,B);
			t2 = get_time();
			t+=(t2-t1);
			A[i%A.size()]+=2.0;
			B[i%B.size()]+=2.0;
			if (i%C.size()==0)
				check+=C[i%B.size()];
		}	

	//cout<<"Timing of "<<print<<" : ";
	//cout.width(10);cout.precision(4);cout.setf(ios::fixed,ios::floatfield);
	double moyenne = t/(double) loop;
	cout<<moyenne<<":";
	//cout<<" ("<<check<<")"<<endl;
}
#define PREC 6
#define SPC2 20
#define TIMING(x) cout<<setw(SPC2)<<std::fixed << std::setprecision(PREC)<<x<<" |";

 template<typename T>
void bench(void mul(vector<T>&,const vector<T>&, const vector<T>&), vector<T> &C, vector<T> A, vector<T> B, size_t loop, string print){
	double t1,t2,t,et1,et2,cst;
	t=0.0;
	double check=0.0;
	cst=2.0;
	et1=get_time();
	for (size_t i= 0; i< loop; i++) 
		{		
			t1 = get_time(); 
			mul(C,A,B);
			t2 = get_time();
			t+=(t2-t1);
			A[A.size()-1]+=cst;
			B[B.size()-1]+=cst;
			if (i%C.size()==0)
				check+=C[i%B.size()];
			cst=-cst;
		}	
	et2=get_time();
	double moyenne = (et2 - et1)/(double) loop;
	TIMING(moyenne);	
}

fft_planner* fft_launcher::fft_plan;
dct_planner* dct_launcher::dct_plan;
fft_planner* dct_via_fft_launcher::fft_plan;

int main(int argc, char** argv)
{ 
	typedef double TYPE;  	 
	size_t degree=4;
	size_t loop=(10000/degree)*(20000/degree);
	srand(time(NULL));
	if (argc == 1){
		cout<<"Usage: [degree] [loop] [seed] \n";
		cout<<"      degree -> degree of polynomial input (DEFAULT="<<degree<<")\n";
		cout<<"      loop   -> nbr of repetition for average timing (DEFAULT="<<loop<<")\n";
		cout<<"      seed   -> seed for random generation of coefficients (DEFAULT=time)\n";
		cout<<endl<<endl;
	}
	if (argc>1)
		degree=atoi(argv[1]);
	if (argc>2)
		loop=atoi(argv[2]);
	if (argc>3)
		srand(atoi(argv[2]));

	//cout<<"number of test: "<<loop<<endl;
	vector<TYPE> A(degree), B(degree), D(degree,0.0), C1(2*degree-1), C2(2*degree-1), C3(2*degree-1), C4(2*degree-1),C5(2*degree-1);//,C6(2*degree-1),C7(2*degree-1),C8(2*degree-1),C9(2*degree-1);
	D[0]=0.0;
	D[1]=1.0;

	
	for (size_t i=0;i<degree;i++){
		A[i]=rand_bound(MAX_V);
		B[i]=rand_bound(MAX_V);
	}

	// create plan for fft and dct
	size_t N=2*degree;
	fft_launcher::fft_plan= new fft_planner(N);
	dct_launcher::dct_plan= new dct_planner(N+1);
	dct_via_fft_launcher::fft_plan= new fft_planner(2*N);


	cout<<"*** Benchmarking Chebyshev Polynomial multiplication (time in seconds) ***\n";
	
	TIMING("degree");
	TIMING("quadratic Cheb.");
	TIMING("DCT Cheb.");
	TIMING("DCT Cheb. FFT");
	TIMING("Karatsuba");
	TIMING("FFT");
	cout<<endl<<endl;;
	TIMING(N);
	bench( cheby_mul,                                  C1, A, B,loop, "Quadratic Chebyshev ");
	bench( dct_launcher::cheby_mul_DCT,                C2, A, B,loop, "DCT based Chebyshev ");
	bench( dct_via_fft_launcher::cheby_mul_DCT,        C3, A, B,loop, "DCT based Cheby  FFT");
	//bench((cheby_mul_red<TYPE,naive_mul>),             C3, A, B,loop, "Quadratic reduction ");
	bench((cheby_mul_red<TYPE,kara_mul2>),             C4, A, B,loop, "Karatsub2 reduction ");
	//bench((cheby_mul_red<TYPE,kara_mul>),              C4, A, B,loop, "Karatsuba reduction ");
	//bench((cheby_mul_red<TYPE,fft_launcher::fft_mul>), C5, A, B,loop, "FFT based reduction ");
	//bench((fft_launcher::cheby_mul_FFT),               C6, A, B,loop, "FFT based red optim ");
	//bench((fft_launcher::cheby_mul_FFT_V2),            C7, A, B,loop, "FFT based red optim2");
	bench((fft_launcher::cheby_mul_FFT_V3),            C5, A, B,loop, "FFT based red optim3");

	cout<<endl;

	
	delete fft_launcher::fft_plan;
	delete dct_launcher::dct_plan;
	delete dct_via_fft_launcher::fft_plan;
	return 0;
}
